module RECTGEN (V_SYNC, H_SYNC, BG_CH, BG_EN, RESET, EN, X, Y, X1, Y1, X2, Y2, CLK, R, G, B) ;

input V_SYNC, H_SYNC;
input RESET;
input CLK;
input [9:0]X;
input [9:0]Y;
input [9:0]X1;
input [9:0]Y1;
input [9:0]X2;
input [9:0]Y2;
input EN;
input BG_CH;
input BG_EN;
output [1:0]R, G, B;

reg [19:0]ball_pos;
reg [19:0]new_ball_pos;
reg [4:0]X_vel;
reg [4:0]Y_vel;
reg [5:0]state;
reg [5:0]BLOCK_COLOR;
reg [9:0]exist;
reg [9:0]exist1;
reg [9:0]exist2;
reg [9:0]exist3;
reg [9:0]exist4;
reg [9:0]exist5;
reg [9:0]diff_ball;
reg block_area;
reg [3:0]color_offset;
reg [2:0]current_row;
reg [3:0]current_block;
reg [9:0]row_exist;
reg block_exist;
reg hit;
reg same_frame;
reg same_row;
reg [5:0]color_cycle;
reg [1:0]bg_select;
reg [1:0]level;

wire PADDLE_EN;
wire BALL_EN;
wire [5:0]color;
wire [5:0]background;
wire [5:0]background1;
wire [5:0]background2;
wire [5:0]background3a;
wire [5:0]background3b;
wire [5:0]background3;
wire [5:0]background4;

assign BALL_EN = ((X >= ball_pos[9:0]) && (X <= ball_pos[9:0] + 5) && (Y >= ball_pos[19:10]) && (Y <= ball_pos[19:10] + 5));
assign PADDLE_EN = ((X >= X1) && (X <= X2) && (Y >= Y1) && (Y <= Y2));
assign color = (PADDLE_EN) ? 3 : ((BLOCK_COLOR) ? BLOCK_COLOR : 'h3F);
assign background1 = Y[5:0] - color_cycle;
assign background2 = X[5:0] + Y[5:0] - color_cycle;
assign background3a = X[5:0] - Y[5:0] + color_cycle;
assign background3b = X[5:0] + Y[5:0] - color_cycle;
assign background3[5] = background3a[5];
assign background3[4] = background3b[4];
assign background3[3] = background3a[3];
assign background3[2] = background3b[2];
assign background3[1] = background3a[1];
assign background3[0] = background3b[0];
assign background4 = (X >> Y[2:0]) - color_cycle;
assign background = (bg_select[1]) ? ((bg_select[0]) ? background4 : background3) : ((bg_select[0]) ? background2 : background1);
assign R = (EN) ? ((PADDLE_EN || BALL_EN || BLOCK_COLOR) ? color[5:4] : ((BG_EN) ? background[5:4] : 0)) : 0;
assign G = (EN) ? ((PADDLE_EN || BALL_EN || BLOCK_COLOR) ? color[3:2] : ((BG_EN) ? background[3:2] : 0)) : 0;
assign B = (EN) ? ((PADDLE_EN || BALL_EN || BLOCK_COLOR) ? color[1:0] : ((BG_EN) ? background[1:0] : 0)) : 0;

always @(posedge CLK) begin
  if(RESET) begin
    ball_pos[9:0] <= 220;
    ball_pos[19:10] <= 240;
    X_vel <= 'h14;
    Y_vel <= 'h12;
    state <= 0;
    exist1 <= 'h3FF;
    exist2 <= 'h3FF;
    exist3 <= 'h3FF;
    exist4 <= 'h3FF;
    exist5 <= 'h3FF;
    block_area <= 0;
    current_row <= 0;
    current_block <= 0;
    row_exist <= 0;
    block_exist <= 0;
    hit <= 0;
    same_frame <= 0;
    same_row <= 0;
    color_cycle <= 0;
    bg_select <= 0;
    level <= 0;
  end

  if (!RESET) begin
      if(BG_CH) begin
        bg_select <= bg_select + 1;
      end
  
      // State 0
      if(!state) begin
        if (V_SYNC) begin
          state <= 1;
          color_cycle <= color_cycle + 1;
          if(!exist1 && !exist2 && !exist3 && ! exist4 && !exist5) begin
            ball_pos[9:0] <= 220;
            ball_pos[19:10] <= 240;
            X_vel <= 'h14;
            Y_vel <= 'h12;
            if(!level) begin
              exist1 <= 'h2AA;
              exist2 <= 'h155;
              exist3 <= 'h2AA;
              exist4 <= 'h155;
              exist5 <= 'h2AA;
              bg_select <= 1;
              level <= 1;
            end
            if(level == 1) begin
              exist1 <= 'h303;
              exist2 <= 'h3CF;
              exist3 <= 'h3FF;
              exist4 <= 'h3CF;
              exist5 <= 'h303;
              bg_select <= 2;
              level <= 2;
            end
            if(level == 2) begin
              exist1 <= 'h231;
              exist2 <= 'h14A;
              exist3 <= 'h084;
              exist4 <= 'h14A;
              exist5 <= 'h231;
              bg_select <= 3;
              level <= 3;
            end
            if(level == 3) begin
              exist1 <= 'h3FF;
              exist2 <= 'h3FF;
              exist3 <= 'h3FF;
              exist4 <= 'h3FF;
              exist5 <= 'h3FF;
              bg_select <= 0;
              level <= 0;
            end
          end
        end
      end

      // State 1
      if(state == 1) begin
        if(V_SYNC) begin
          block_area <= 0;
          if(Y[9:4] == 2) begin
            exist <= exist1;
            block_area <= 1;
            color_offset <= 0;
          end
          if(Y[9:4] == 3) begin
            exist <= exist2;
            block_area <= 1;
            color_offset <= 5;
          end
          if(Y[9:4] == 4) begin
            exist <= exist3;
            block_area <= 1;
            color_offset <= 10;
          end
          if(Y[9:4] == 5) begin
            exist <= exist4;
            block_area <= 1;
            color_offset <= 15;
          end
          if(Y[9:4] == 6) begin
            exist <= exist5;
            block_area <= 1;
            color_offset <= 20;
          end
          if(Y[3:0] == 'hF) begin
            block_area <= 0;
          end
        end
        state <= 2;
        if (!V_SYNC)
          state <= 3;
      end

      // State 2
      if(state == 2) begin
        BLOCK_COLOR <= 0;
        if(block_area) begin
          if(!((X + 1) >> 6)) begin
            BLOCK_COLOR <= (exist[0]) ? 1 + color_offset : 0;
          end
          if((X + 1) >> 6 == 1) begin
            BLOCK_COLOR <= (exist[1]) ? 2 + color_offset : 0;
          end
          if((X + 1) >> 6 == 2) begin
            BLOCK_COLOR <= (exist[2]) ? 3 + color_offset : 0;
          end
          if((X + 1) >> 6 == 3) begin
            BLOCK_COLOR <= (exist[3]) ? 4 + color_offset : 0;
          end
          if((X + 1) >> 6 == 4) begin
            BLOCK_COLOR <= (exist[4]) ? 5 + color_offset : 0;
          end
          if((X + 1) >> 6 == 5) begin
            BLOCK_COLOR <= (exist[5]) ? 6 + color_offset : 0;
          end
          if((X + 1) >> 6 == 6) begin
            BLOCK_COLOR <= (exist[6]) ? 7 + color_offset : 0;
          end
          if((X + 1) >> 6 == 7) begin
            BLOCK_COLOR <= (exist[7]) ? 8 + color_offset : 0;
          end
          if((X + 1) >> 6 == 8) begin
            BLOCK_COLOR <= (exist[8]) ? 9 + color_offset : 0;
          end
          if((X + 1) >> 6 == 9) begin
            BLOCK_COLOR <= (exist[9]) ? 10 + color_offset : 0;
          end
          if((X + 1) >> 6 > 12) begin
            BLOCK_COLOR <= (exist[0]) ? 1 + color_offset : 0;
          end
          if((X[5:0] + 1) == 'h3F) begin
            BLOCK_COLOR <= 0;
          end
        end
        state <= (H_SYNC) ? 2 : ((!V_SYNC) ? 3 : 1);  
      end

      // State 3         
      if(state == 3) begin
        new_ball_pos[9:0] <= (X_vel[4]) ? ball_pos[9:0] - X_vel[3:0] : ball_pos[9:0] + X_vel[3:0];    
        new_ball_pos[19:10] <= (Y_vel[4]) ? ball_pos[19:10] - Y_vel[3:0] : ball_pos[19:10] + Y_vel[3:0];
        diff_ball <= (ball_pos[9:0] - X1);
        state <= (ball_pos[19:10] < 300) ? 14 : 6;
        same_frame <= 0;
        hit <= 0;
      end

      if(state == 4) begin
        state <= 5;
        ball_pos[9:0] <= (new_ball_pos[9:0] >= 635 && new_ball_pos[9:0] < 645) ? 635 : new_ball_pos[9:0];
        X_vel[4] <= (new_ball_pos[9:0] >= 635 && new_ball_pos[9:0] < 645) ? 1 : X_vel[4];
        if(new_ball_pos[19:10] >= 475 && new_ball_pos[19:10] <= 485) begin
          ball_pos[9:0] <= 220;
          ball_pos[19:10] <= 240;
          X_vel <= 'h14;
          Y_vel <= 'h12;
          state <= 0;
        end else begin
          ball_pos[19:10] <= new_ball_pos[19:10];
        end
      end
      
      if(state == 5) begin
        ball_pos[9:0] <= (!new_ball_pos[9:0] || (new_ball_pos[9] && new_ball_pos[8])) ? 0 : ball_pos[9:0];
        X_vel[4] <= (!new_ball_pos[9:0] || (new_ball_pos[9] && new_ball_pos[8])) ? 0 : X_vel[4];
             
        ball_pos[19:10] <= (!new_ball_pos[19:10] || (new_ball_pos[19] && new_ball_pos[18])) ? 0 : new_ball_pos[19:10];
        Y_vel[4] <= (!new_ball_pos[19:10] || (new_ball_pos[19] && new_ball_pos[18])) ? 0 : Y_vel[4];
             
        state <= 0;
      end
         
      // Check to see if ball is in valid X region
      if(state == 6) begin
        if(X1)
          state <= ((ball_pos[9:0] <= X2) && (ball_pos[9:0] >= X1 - 5)) ? 7 : 4;
        else
          state <= ((ball_pos[9:0] <= X2)  && ball_pos[9:0] >= 0) ? 7 : 4;
      end
         
      // Check to see if ball is in valid Y region
      if(state == 7)begin
        state <= ((ball_pos[19:10] <= Y2) && (ball_pos[19:10] >= Y1 - 5)) ? 8 : 4;
      end

      // Check to see if ball is fully on top of paddle         
      if(state == 8) begin
        state <= ((ball_pos[9:0] <= X2 - 5) && (ball_pos[9:0] >= X1)) ? 10 : 9;
      end
                  
      // Check to see if ball is on side of paddle
      if(state == 9) begin
           state <= ((ball_pos[19:10] >= Y1) && (ball_pos[19:10] <= Y2)) ? 12 : 13;
      end
      
      // In this state, the ball is above the paddle, so bounce.
      if(state == 10) begin       
        ball_pos[19:10] <= (!Y_vel[4]) ? Y1 - 6 : Y2 + 1;
        Y_vel[4] <= (!Y_vel[4])? 1 : 0;
        state <= 11; 
      end
      
      if(state == 11) begin
        if (diff_ball > 0 && diff_ball < 5) begin
          X_vel <= 'h15;
          Y_vel <= 'h11;
        end
        
        if (diff_ball >= 5 && diff_ball < 16) begin
          X_vel <= 'h14;
          Y_vel <= 'h12;
        end
        
        if (diff_ball >= 16 && diff_ball < 27) begin
          X_vel <= 'h12;
          Y_vel <= 'h14;
        end
        
        if (diff_ball >= 27 && diff_ball < 33) begin
          X_vel <= 'h00;
          Y_vel <= 'h15;
        end
        
        if (diff_ball >= 33 && diff_ball < 44) begin
          X_vel <= 'h02;
          Y_vel <= 'h14;
        end
        
        if (diff_ball >= 44 && diff_ball < 55) begin
          X_vel <= 'h04;
          Y_vel <= 'h12;
        end
        
        if (diff_ball >= 55 && diff_ball < 65) begin
          X_vel <= 'h05;
          Y_vel <= 'h11;
        end
        
        state <= 0;
      end
         
      // In this state, the ball hit the side of the paddle, so bounce.   
      if(state == 12) begin
        ball_pos[9:0] <= (!X_vel[4]) ? X1 - 6 : X2 + 1;
        X_vel[4] <= (!X_vel[4]) ? 1 : 0;
        state <= 0;
      end
      
      // In this state, we have hit the corner of the paddle.         
      if(state == 13) begin
        ball_pos[9:0] <= (ball_pos[9:0] > X1 + 30)  ? X2 + 1 : X1 - 6;
        X_vel[4] <= (ball_pos[9:0] > X1 + 30) ? 0 : 1;
        ball_pos[19:10] <= Y1 - 6;
        Y_vel[4] <= 1;
        X_vel[3:0] <= 5;
        Y_vel[3:0] <= 1;
        state <= 0;
      end
      
      
      // State 14: handle block interaction
      if(state == 14) begin
        same_row <= 0;
        block_exist <= 0;
        state <= 15;
        if(!same_frame) begin
          row_exist <= exist1;
          current_row <= 0;
          same_frame <= 1;
        end
        if(current_row == 0 && same_frame) begin
          exist1 <= row_exist;
          row_exist <= exist2;
          current_row <= 1;
        end
        if(current_row == 1) begin
          exist2 <= row_exist;
          row_exist <= exist3;
          current_row <= 2;
        end
        if(current_row == 2) begin
          exist3 <= row_exist;
          row_exist <= exist4;
          current_row <= 3;
        end
        if(current_row == 3) begin
          exist4 <= row_exist;
          row_exist <= exist5;
          current_row <= 4;
        end
        if(current_row >= 4) begin
          exist5 <= row_exist;
          current_row <= 0;
          state <= (hit) ? 0 : 4;
        end
      end

      if(state == 15) begin
        state <= 16; 
        if(!same_row) begin
          block_exist <= row_exist[0];
          current_block <= 0;
          same_row <= 1;
        end
        if(current_block == 0 && same_row) begin
          row_exist[0] <= block_exist;
          block_exist <= row_exist[1];
          current_block <= 1;
        end
        if(current_block == 1) begin
          row_exist[1] <= block_exist;
          block_exist <= row_exist[2];
          current_block <= 2;
        end
        if(current_block == 2) begin
          row_exist[2] <= block_exist;
          block_exist <= row_exist[3];
          current_block <= 3;
        end
        if(current_block == 3) begin
          row_exist[3] <= block_exist;
          block_exist <= row_exist[4];
          current_block <= 4;
        end
        if(current_block == 4) begin
          row_exist[4] <= block_exist;
          block_exist <= row_exist[5];
          current_block <= 5;
        end
        if(current_block == 5) begin
          row_exist[5] <= block_exist;
          block_exist <= row_exist[6];
          current_block <= 6;
        end
        if(current_block == 6) begin
          row_exist[6] <= block_exist;
          block_exist <= row_exist[7];
          current_block <= 7;
        end
        if(current_block == 7) begin
          row_exist[7] <= block_exist;
          block_exist <= row_exist[8];
          current_block <= 8;
        end
        if(current_block == 8) begin
          row_exist[8] <= block_exist;
          block_exist <= row_exist[9];
          current_block <= 9;
        end
        if(current_block >= 9) begin
          row_exist[9] <= block_exist;
          current_block <= 0;
          state <= 14;
        end
      end
      
      // Check to see if ball is in valid X region
      if(state == 16) begin
        if(current_block)
          state <= ((ball_pos[9:0] <= ((current_block << 6) + 63)) && (ball_pos[9:0] >= (current_block << 6) - 5) && block_exist) ? 17 : 15;
        else
          state <= ((ball_pos[9:0] <= ((current_block << 6) + 63)) && (ball_pos[9:0] >= 0) && block_exist) ? 17 : 15;
      end
         
      // Check to see if ball is in valid Y region
      if(state == 17) begin
        state <= ((ball_pos[19:10] <= (((current_row + 2) << 4) + 15)) && (ball_pos[19:10] >= (((current_row + 2) << 4) - 5))) ? 18 : 15;
      end

      // Check to see if ball is fully on top or bottom of brick
      if(state == 18) begin
        state <= ((ball_pos[9:0] <= ((current_block << 6) + 58)) && (ball_pos[9:0] >= (current_block << 6))) ? 20 : 19;
        block_exist <= 0;
      end
                  
      // Check to see if ball is on side of brick
      if(state == 19) begin
           state <= ((ball_pos[19:10] >= ((current_row + 2) << 4)) && (ball_pos[19:10] <= (((current_row + 2) << 4) + 10))) ? 21 : 22;
      end
      
      // In this state, the ball is above or below the brick, so bounce.
      if(state == 20) begin
        ball_pos[9:0] <= (new_ball_pos[9:0] >= 635) ? 635 : new_ball_pos[9:0];
        X_vel[4] <= (new_ball_pos[9:0] >= 635) ? 1 : X_vel[4];
        
        ball_pos[19:10] <= (!Y_vel[4]) ? ((current_row + 2) << 4) - 6 : ((current_row + 2) << 4) + 16;
        Y_vel[4] <= (!Y_vel[4])? 1 : 0;
        state <= 15;
        hit <= 1;
      end
      
      // In this state, the ball hit the side of the brick, so bounce.   
      if(state == 21) begin
        ball_pos[9:0] <= (!X_vel[4]) ? (current_block << 6) - 6 : (current_block << 6) + 64;
        X_vel[4] <= (!X_vel[4]) ? 1 : 0;
        state <= 15;
        hit <= 1;
      end
      
      // In this state, we have hit the corner of the brick.         
      if(state == 22) begin
        ball_pos[9:0] <= (ball_pos[9:0] > (current_block << 6) + 30)  ? (current_block << 6) + 64 : (current_block << 6) - 6;
        X_vel[4] <= (ball_pos[9:0] > (current_block << 6) + 30) ? 0 : 1;
        ball_pos[19:10] <= (ball_pos[19:10] > ((current_row + 2) << 4) + 4) ? ((current_row + 2) << 4) + 16 : ((current_row + 2) << 4) - 6;
        Y_vel[4] <= (ball_pos[19:10] > ((current_row + 2) << 4) + 4) ? 0 : 1;
        state <= 15;
        hit <= 1;
      end
  end
end


endmodule 
